#define BUFSTAT

#define NCON 20                         /* maximum number of connections */
#define NHOST  20                       /* maximum number of foreign hosts */
#define NGATE	20			/* maximum size of gateway table */
#define NWORK 20                        /* # of work entries in table */

#define TRUE  1
#define FALSE 0

#ifndef HZ
#define HZ 60
#endif
#ifndef NULL
#define NULL 0
#endif

typedef unsigned long unlong;
typedef unsigned char unchar;


typedef unsigned short n_short;	/* short as received from the net */
typedef unlong n_long;   	/* long as received from the net */
typedef unlong sequence; 	/* sequence numbers in tcp */
typedef	unlong net_t;		/* network number */

/* worklist entry */
struct work {
	short	     w_type;	/* what to do with entry */
	short	     w_stype;	/* subtype for timer names */
	struct tcb  *w_tcb;	/* -> tcb for entry */
	char	    *w_dat;	/* -> work data chain */
	struct work *w_next;	/* -> next work entry */
};

/* all purpose internet address */
struct socket {
				/* (can be extended for any network) */
	union {
		struct {
        		unchar s_b1;
        		unchar s_b2;
        		unchar s_b3;
        		unchar s_b4;
		}S_un_b;
		struct {
			unsigned short s_w1;
			unsigned short s_w2;
		}S_un_w;
		unsigned long s_l;
	} S_un;
#define s_addr  S_un.s_l        /* can be used for most tcp & ip code */
};

/* net representation to right justified long integer */
#define netoint(x) (int)(x)

/* make an ip addr */
net_t ipaddr();

/* host table entry */
struct host {
	struct socket   h_addr;		/* internet address */
	unchar          h_rfnm;		/* # outstanding rfnm's */
	unchar          h_refct;	/* reference count */
	unsigned short  h_stat;		/* host status */
	struct mbuf    *h_outq;		/* -> host msg holding queue */
};

#define HUP 	128		/* host up */
#define HTEMP 	64		/* temp host entry (goes away if h_rfnm==0) */

/* network buffer allocation control */
struct net {                   
	struct proc  *n_proc;		/* -> network process */
	struct inode *n_debug;		/* -> inode of debugging file */

 	struct ipq   *n_ip_head;	/* -> top of ip reass. queue */
	struct ipq   *n_ip_tail;	/* -> end of ip reass. queue */
	short         n_ip_lock;	/* ip reass. queue lock */

	struct proto *n_ip_proto;	/* -> ip raw proto chain */
	struct proto *n_udp_proto;	/* -> udp raw proto chain */
	struct proto *n_icmp_proto;	/* -> icmp raw proto chain */
	struct proto *n_ggp_proto;	/* -> ggp raw proto chain */
	struct proto *n_tcp_proto;	/* -> tcp raw proto chain */

	struct work  *n_work;		/* -> top of tcp work queue */
	struct tcb   *n_tcb_head;	/* -> top of tcp tcb list */
	struct tcb   *n_tcb_tail;	/* -> end of tcp tcb list */
	sequence      n_iss;		/* tcp initial send seq # */

	short n_free;			/* index of top of page free list */
#define netlow n_free			/* page # of lowest mem used by net (mbb) */
	short n_pages;			/* # pages owned by network */
	short n_bufs;			/* # free msg buffers remaining */
	short n_hiwat;			/* max. # free mbufs allocated */
	short n_lowat;			/* min. # free mbufs allocated */
	unsigned short n_ip_cnt;	/* ip packet counter, used for ids */

	short n_logct;			/* logging queue size */
	short n_loglim;			/* logging queue limit */
	struct mbuf *n_log_hd;		/* logging queue head */
	struct mbuf *n_log_tl;		/* logging queue tail */
};

struct net_stat {               /* interesting statistics */
	int imp_resets;                 /* # times imp reset */
	int imp_flushes;                /* #packets flushed by imp */
	int imp_drops;                  /* #msgs from imp no-one wants */
	int m_drops;                    /* #mbuf drops from lack of bufs */
	int ip_badsum;                  /* #bad ip checksums */
	int ip_drops;			/* #ip packets not addressed to us */
	int t_badsum;                   /* #bad tcp checksums */
	int t_badsegs;                  /* #bad tcp segments */
	int t_unack;                    /* #tcp segs placed on rcv_unack */
	int ic_drops;			/* #bad icmp messages */
	int ic_badsum;			/* #bad icmp checksums */
	int ic_quenches;		/* #icmp source quenches */
	int ic_redirects;		/* #icmp redirects */
	int ic_echoes;			/* #icmp echoes */
	int ic_timex;			/* #icmp time exceeded messages */
	int u_badsum;			/* #udp bad checksums */
	int u_drops;			/* #udb drops */
};

struct buf_stat {		/* buffer usage statistics */
	int b_cons;			/* active connections */
	int b_useq;			/* user send+rcv queues */
	int b_work;			/* in transit on work list */
	int b_devq;			/* device input/output queues */
	int b_rfnm;			/* blocked for rfnm */
	int b_ipfr;			/* ip fragment reassembly queues */
	int b_tcbs;			/* tcbs */
	int b_tseq;			/* tcp sequencing queues */
	int b_tuna;			/* tcp unacked message queues */
};

struct t_debug {                /* tcp debugging record */
        long t_tod;                     /* time of day */
	struct tcb *t_tcb;              /* -> tcb */
	char t_old;                     /* old state */
	char t_inp;                     /* input */
	char t_tim;                     /* timer id */ 
	char t_new;                     /* new state */
	sequence t_sno;                 /* sequence number */
	sequence t_ano;                 /* acknowledgement */
	unsigned short t_wno;           /* window */
	unsigned short t_lno;           /* length */
	unchar t_flg;			/* message flags */
};

#ifdef KERNEL

#define MIN(a, b) ((a)<(b) ? (a) : (b))

typedef struct socket netaddr;
struct work *workfree, *workNWORK;      /* ->work entry table */
int nwork;
struct host *host, *hostNHOST;          /* ->start, end of host table */
int nhost;
extern struct net netcb;                /* network control block */
extern struct net_stat netstat;         /* net statistics block */
#ifdef BUFSTAT
extern struct buf_stat bufstat;		/* buffer stats */
#endif /* BUFSTAT */

#endif /* KERNEL */
