#include <stdio.h>

#include "../include/param.h"
#include "../include/inode.h"
#include "mbuf.h"
#include "net.h"
#include "ifcb.h"
#include "tcp.h"
#include "ip.h"
#include "ucb.h"
#include "fsm.h"
#include "../include/buf.h"
#include "../include/user.h"
#include "../include/proc.h"

#define MINOR(d) (((d) >> 8) & 0xFF)
#define MAJOR(d) ((d) & 0xFF)

#define iptonet(a) (a)

extern struct user u;
char netbugf[] = "/etc/net/tcpdebug";
char gatefile[] = "/etc/net/gateway.bin";

/*
 * This is the mainline of the network input process.  First, we
 * fork off as a separate process within the kernel, initialize the
 * net buffer manager, local net interfaces, and timers.  Then, we
 * loop forever, sleeping on net input.  Once awakened by a local net
 * driver, we dispatch to each local interface in turn, to process the
 * input message(s).  After handling the interfaces, we adjust the net
 * buffer allocation, call on the tcp processor to handle any outstanding
 * timer or user events, and...
 */
main()
{
	register struct inode *ip;
	register struct proc *p;
	register struct ifcb *lp;
	int schar();
	int fid;

#ifdef FORK
	if ((fid = fork()) == 0) {
#endif

		meminit(); /*mbufinit();*/     /* init buf mgmt system */
		gatinit();	/* read gateway file and init table */

		/* initialize all local net i/fs */

		for (lp=ifcb; lp < &ifcb[nifcb]; lp++) { 
			(*lp->if_init)(lp);
		}

		/* start timers */

		net_timer();

		for (;;) {      /* body of input process */

			/*
			 * reset local i/f if necessary, otherwise
			 * just handle net input
			 */
			for (lp=ifcb; lp < &ifcb[nifcb]; lp++) {
process:
				if (lp->if_error && !lp->if_needinit) {
					lp->if_needinit = TRUE;
					netreset(lp);
				} else if (lp->if_avail && lp->if_inq_hd != NULL)
					(*lp->if_rcv)(lp);
			
			/* call TCP processor to handle any user or
			   timer events. */
				
				if (netcb.n_work != NULL)
					tcp_input(NULL, NULL);
				
			/*
			 * expand or shrink the buffer freelist as needed
			 */
	        		if (netcb.n_bufs < netcb.n_lowat)
	        			m_expand();
	        		else if (netcb.n_bufs > netcb.n_hiwat)
	        			m_relse();
			}
	
			/* 
			 * wait for net input (awakened by drivers)
			 */
			for (lp=ifcb; lp < &ifcb[nifcb]; lp++) {
				(*lp->if_poll)(lp);
				if (lp->if_inq_hd != NULL)
					goto process;
			}
			ksleep((caddr_t)ifcb, PZERO-15);
		}
#ifdef FORK
	}
#endif
}

#ifdef BUFSTAT
struct buf_stat bufnull = {0};
#endif /* BUFSTAT */

/*
 * Net timer routine: scheduled every second, calls ip and tcp timeout routines
 * and does any net-wide statistics.
 */
net_timer()
{
	register struct mbuf *m, *n;
	register struct ucb *up;
	register struct tcb *tp;
	register struct th *t;
	register struct ifcb *lp;
	register struct ipq *ip;
	register struct ip *i;
	register struct host *h;
	register struct work *w;

	ip_timeo();
	tcp_timeo();

#if 0
#ifdef BUFSTAT
	bufstat = bufnull;
	for (up=contab; up < conNCON; up++) {
		if (up->uc_proc == NULL)
			continue;
		if (up->uc_flags & UTCP) {
			if ((tp = up->uc_tcb) == NULL)
				continue;
			bufstat.b_tcbs++;
			if (tp->t_state != ESTAB)
				bufstat.b_cons--;
			if ((t = tp->t_rcv_next) == NULL)
				continue;
			for (; t != (struct th *)tp; t=t->t_next) {
				m = dtom(t);
				while (m != NULL) {
					bufstat.b_tseq++;
					m = m->m_next;
				}
			}
			for (m=tp->t_rcv_unack; m != NULL; m=m->m_act) {
				n = m;
				do {
					bufstat.b_tuna++;
					n = n->m_next;
				} while (n != NULL);
			}
		}
		bufstat.b_cons++;
		bufstat.b_useq += up->uc_rsize + up->uc_ssize;
	}
	for (ip=netcb.n_ip_head; ip != NULL; ip=ip->iq_next) 
		for (i=ip->iqx.ip_next; i != (struct ip *)ip; i=i->ip_next) {
			m = dtom(i);
			while (m != NULL) {
				bufstat.b_ipfr++;
				m = m->m_next;
			} 
		}
	for (lp=ifcb; lp < &ifcb[nifcb]; lp++) {
		for (m=lp->if_outq_hd; m != NULL; m = m->m_act) {
			n = m;
			do {
				bufstat.b_devq++;
				n = n->m_next;
			} while (n != NULL);
		}
		for (m=lp->if_outq_cur; m != NULL; m = m->m_next)
			bufstat.b_devq++;
		for (m=lp->if_inq_hd; m != NULL; m = m->m_act) {
			n = m;
			do {
				bufstat.b_devq++;
				n = n->m_next;
			} while (n != NULL);
		}
		for (m=lp->if_inq_msg; m != NULL; m = m->m_next)
			bufstat.b_devq++;
	}
	for (h=host; h < hostNHOST; h++)
		if (h->h_refct > 0) 
			for (m = h->h_outq; m != NULL; m = m->m_act) {
				n = m;
				do {
					bufstat.b_rfnm++;
					n = n->m_next;	
				} while (n != NULL);
			}
	for (w=netcb.n_work; w != NULL; w = w->w_next)
		if (w->w_type == INRECV || w->w_type == IUSEND)
			for (m=dtom(w->w_dat); m != NULL; m = m->m_next) 
				bufstat.b_work++;
#endif /* BUFSTAT */

	timeout(net_timer, 0, HZ);
#endif
}

/*
 * Generic local net reset:  clear input and output queues and call driver
 * init routine.
 */
netreset(ip)
register struct ifcb *ip;
{
	register struct mbuf *m, *n;
	int s;	

	s = spl5();
	ip->if_avail = FALSE;

	/*
	 * clear all i/f queues
	 */
	for (m=ip->if_inq_hd; m != NULL; m=n) {
		n = m->m_act;
		m_freem(m);
	}
	ip->if_inq_hd = NULL;
	ip->if_inq_tl = NULL;

	for (m=ip->if_outq_hd; m != NULL; m=n) {
		n = m->m_act;
		m_freem(m);
	}
	ip->if_outq_hd = NULL;
	ip->if_outq_tl = NULL;

	if (ip->if_inq_msg != NULL) {
		m_freem(ip->if_inq_msg);
		ip->if_inq_msg = NULL;
		ip->if_inq_cur = NULL;
	}

	if (ip->if_outq_cur != NULL) {
		m_freem(ip->if_outq_cur);
		ip->if_outq_cur = NULL;
	}

	h_reset(iptonet(ip->if_addr));
	splx(s);
	
	/* 
	 * now call net driver specific init routine to reinit
	 */
	(*ip->if_init)(ip);
}

/*
 * Read the binary format gateway file into the gateway table.  The file
 * contains gateway structure entries produced from the ASCII gateway
 * table.  Called from netmain or netioctl, returns number of entries
 * successfully read or -1 if error
 */
gatinit()
{
#if 0
	register i;
	int fd;
	register struct ifcb *lp;
	register struct gway *gp;
	struct gway gatent;
	net_t gatnet;
	int schar();

	/* set up to read the gateway file */

	if ((fd = open(gatefile, O_RDONLY)) < 0) 
		return(-1);

	/* read entries from the file into the table */

	for (gp=gateway, i=NGATE; i > 0; i--, gp++) {
		u.u_base = (caddr_t)&gatent;
		u.u_count = sizeof gatent;
		read (fd, );
		if (u.u_error != 0) {
			printf("error %d reading gateway file: %s\n",
				u.u_error, gatefile);
			goto out;
		}
		if (u.u_count == sizeof gatent) 	/* check EOF */
			break;
		if (u.u_count != 0 || !gatent.g_flags & GWFORMAT) {
			printf("invalid gateway file entry, read %d entries\n",
				NGATE-i+1);
			goto out;
		}

		/* look for i/f corresponding to gateway local net addr */

		gatnet = iptonet(gatent.g_local);
		for (lp=ifcb; lp < &ifcb[nifcb]; lp++)
			if (iptonet(lp->if_addr) == gatnet) {
				gatent.g_ifcb = lp;
				goto gotif;
			}
		printf("invalid gateway addr %X, entry %d ignored\n",
			gatent.g_local.s_addr, NGATE-i+1);
		gp--;
gotif:
		*gp = gatent;
	}
	if (u.u_offset < ip->i_size)
		tablefull("gateway");
out:
	gatewayNGATE = gp;
	close(fd);
	return(NGATE-i+1);
#endif
}

/*
 * Save anomolous packets for debugging purposes.  If limit is non-zero,
 * save up to limit packets.  Otherwise just free.
 */
netlog(mp)
register struct mbuf *mp;
{
#if 0
	register struct mbuf *m;
		
	if (netcb.n_loglim > 0) {

		/* free all but header */

		if (mp->m_next != NULL) {
			m_freem(mp->m_next);
			mp->m_next = NULL;
		}

		/* add to end */

		mp->m_act = NULL;
		if (netcb.n_log_hd != NULL)
			netcb.n_log_tl->m_act = mp;
		else
			netcb.n_log_hd = mp;
		netcb.n_log_tl = mp;
		netcb.n_logct++;
		
		/* if limit exceeded, remove from head */

		while (netcb.n_logct > netcb.n_loglim && 
		    (m = netcb.n_log_hd) != NULL) {
			netcb.n_log_hd = m->m_act;
			m_free(m);
			netcb.n_logct--;
		}
	} else
#endif
		m_freem(mp);
}

/* OS stubs */

#include "../include/file.h"
#include "../include/reg.h"

void
iput()
{
	return;
}

int
suser()
{
	return 1;
}

void
wakeup()
{
	return;
}

void
ksleep()
{
	sleep (1);
	return;
}

void
psignal()
{
	return;
}

int
copyin(dst, src, cnt)
	char *dst, *src;
	int cnt;
{
	memcpy( dst, src, cnt);
	return cnt;
}

int
copyout(dst, src, cnt)
	char *dst, *src;
	int cnt;
{
	memcpy( dst, src, cnt);
	return cnt;
}

void
iomove(kdata, n, flag)
	char *kdata;
	register int n, flag;
{
	if (u.u_segflg) {
		/* copy within kernel space */
		if (flag==B_WRITE)
			memcpy(kdata, u.u_base, n);
		else
			memcpy(u.u_base, kdata, n);
	} else {
		/* copy a block from/to user space */
		if (flag==B_WRITE)
			copyin(kdata, u.u_base, n);
		else
			copyout(u.u_base, kdata, n);		
	}
	u.u_base   += n;
	u.u_offset += n;
	u.u_count  -= n;
}


int
ufalloc()
{
	register int i;

	for (i=0; i<NOFILE; i++)
		if (u.u_ofile[i] == NULL) {
			u.u_ar0[R0] = i;
			return(i);
		}
	u.u_error = EMFILE;
	return(-1);
}

struct file *
falloc()
{
	register struct file *fp;
	register int i;

	i = ufalloc();
	if (i < 0)
		return(NULL);
	for (fp = &file[0]; fp < &file[NFILE]; fp++)
		if (fp->f_count==0) {
			u.u_ofile[i] = fp;
			fp->f_count++;
			fp->f_offset = 0;
			return(fp);
		}
	printf("no file\n");
	u.u_error = ENFILE;
	return(NULL);
}

int
spl0()
{
	return 0;
}


int
spl4()
{
	return 0;
}

int
spl5()
{
	return 0;
}

int
splx()
{
	return 0;
}

void
m_expand()
{
	return;
}

void
m_relse()
{
	return;
}
