
#include "../include/param.h"
#include "../include/user.h"

#include "net.h"
#include "mbuf.h"
#include "ip.h"
#include "ucb.h"
#include "udp.h"
#include "ifcb.h"

extern struct user u;
extern int nosum;

/*
 *  Process incoming udb messages.  Called directly from ip_input.
 */
udp_input(mp, ip)
	register struct mbuf *mp;
	struct ifcb *ip;
{
	register struct udp *p;
	register struct ucb *up;
	register unsigned short lport, fport, i, ulen;

	/* Fix up IP header into UDP pseudo header */
	p = (struct udp *)((int)mp + mp->m_off);
	mp->m_off += sizeof p->u_x;
	mp->m_len -= sizeof p->u_x;
	p->u_x1    = 0;
	p->u_ilen  = p->u_len;

	ulen  = p->u_len;
	lport = p->u_dst;
	fport = p->u_src;
	
	/* Do checksum calculation */
	if ((i = p->u_sum) != 0) {
		p->u_sum = 0;
		if (i != (unsigned short)cksum(mp, ulen + UDPCKSIZE)) {
			netstat.u_badsum++;
			if (nosum)
				goto ignore;
			netlog(mp);
			return;
		}
	} 

ignore:
			
	/* find a ucb for incoming message (exact match) */

	for (up=contab;	up < conNCON && 
		   (up->uc_proc == NULL ||
		    !(up->uc_flags & UUDP) ||
		    up->uc_udp.u_lport != lport ||
		    up->uc_udp.u_fport != fport ||
		    up->uc_host.s_addr != p->u_s.s_addr ||
		    up->uc_local.s_addr!= p->u_d.s_addr); up++);
	
	/* if no exact match, try wild card match */

	if (up == conNCON)
		for (up=contab; up < conNCON &&
			(up->uc_proc == NULL ||
			 !(up->uc_flags & UUDP) ||
			 up->uc_udp.u_lport != lport ||
			 (up->uc_udp.u_fport != fport && 
			  up->uc_udp.u_fport != 0) ||
			 (up->uc_host.s_addr != p->u_s.s_addr &&
			  up->uc_host.s_addr != 0) ||
			 (up->uc_local.s_addr != p->u_d.s_addr &&
			  up->uc_local.s_addr != 0)); up++);
		
	/* if a user is found, queue the data, otherwise drop it */

	if (up != conNCON) {
		mp->m_off += (UDPSIZE + UDPCKSIZE);
		mp->m_len -= (UDPSIZE + UDPCKSIZE);
		raw_queue(mp, up);
	} else {
		netstat.u_drops++;
		m_freem(mp);
	}
}

/*
 * Output a udp message.  Called from netuser in the manner of the raw
 * message interface.  Passed an mbuf chain whose first buffer is empty
 * and used for leader construction.
 */
udp_output(up, mp, len)
struct ucb *up;
register struct mbuf *mp;
int len;
{
	register struct udp *p;
	register struct mbuf *m;

	/* construct header at bottom of first mbuf */
	mp->m_off = MSIZE - sizeof(struct udp);
	mp->m_len = sizeof(struct udp);
	p = (struct udp *)((int)mp + mp->m_off);

	/* fill in header fields */
	p->u_len = len+UDPSIZE;

	p->u_src = up->uc_udp.u_lport;
	p->u_dst = up->uc_udp.u_fport;

	/* Fill in pseudo-header fields */
	p->u_x1   = 0;
	p->u_pr   = UDPROTO;
	p->u_ilen = p->u_len;			/* redundant! */
	p->u_s    = up->uc_local;
	p->u_d    = up->uc_host;

	/* checksum calculation */
	mp->m_off += sizeof p->u_x;
	mp->m_len -= sizeof p->u_x;
	p->u_sum   = 0;
	p->u_sum   = cksum(mp, len + sizeof(struct udp) - sizeof p->u_x);
	mp->m_off -= sizeof p->u_x;
	mp->m_len += sizeof p->u_x;

	/*
	 * Now send the packet via IP.
	 */
	if (!ip_output(up, mp, UDPROTO, len+UDPSIZE, 0, NULL))
		u.u_error = ERAWBAD;
}
