#include <stdio.h>
#include <errno.h>
#include "inet.h"

#include "net.h"
#include "ucb.h"
#include "tcp.h"
#include "ifcb.h"

static struct ucb ucb;
static struct tcb tcb;
static struct ifcb lifcb;

static unsigned long
nettoip (ip)
unsigned long ip;
{
   if (ip != 0L)
   {
      while ((ip & 0xFF000000L) == 0L)
	 ip <<= 8;
   }
   return ip;
}

main (argc, argv)
     int argc;
     char *argv[];
{
   struct ucb *up;
   struct ifcb *lp;
   struct gway *gp;
   register i, j, tcp, udp;
   char all, opts, rstats, istats, pstats, mstats, cstats;

   all = opts = istats = pstats = mstats = rstats = cstats = 0;

   while (argc > 1 && argv[1][0] == '-')
   {
      switch (argv[1][1])
      {

      case 'a':
         cstats++;
         istats++;
         mstats++;
         pstats++;
         rstats++;
	 all++;
         break;

      case 'c':
         cstats++;
         break;

      case 'i':
         istats++;
         opts++;
         break;

      case 'm':
         mstats++;
         opts++;
         break;

      case 'r':
         rstats++;
         opts++;
         break;

      case 's':
         pstats++;
         opts++;
         break;

      default:
       bad:
         printf ("usage: %s [-acimrs] \n", argv[0]);
         return;
      }
      argv++;
      argc--;
   }

   if (!opts || cstats)
   {
      printf (
"USER      PID  MODE  STAT FOREIGN HOST    FPRT  LPRT  TCB   SB RB SA RA TS\n");

      for (up = netcb.n_ucb_hd; up != NULL; up = up->uc_next)
      {

         memcpy (&ucb, up, sizeof (struct ucb));
         if (ucb.uc_proc != 0)
         {

            printf ("         ");

            printf (" %4X ", ucb.uc_proc);

            i = 5;
            tcp = 0;
	    udp = 0;
            if (ucb.uc_flags & UTCP)
            {
               tcp++;
               putchar ('T');
	       if (ucb.uc_tcb != 0)
		  memcpy (&tcb, ucb.uc_tcb, sizeof tcb);
            }
            else if (ucb.uc_flags & UIP)
               putchar ('I');
            else if (ucb.uc_flags & URAW)
               putchar ('R');
            else if (ucb.uc_flags & UCTL)
               putchar ('C');
            else if (ucb.uc_flags & UUDP)
	    {
               putchar ('U');
	       udp++;
	    }
            else
               putchar ('?');
            if (ucb.uc_flags & UEOL)
            {
               putchar ('P');
               i--;
            }
            if (ucb.uc_flags & UURG)
            {
               putchar ('U');
               i--;
            }
            if (ucb.uc_flags & RAWERR)
            {
               putchar ('E');
               i--;
            }
            if (ucb.uc_flags & RAWASIS)
            {
               putchar ('A');
               i--;
            }
            else if (ucb.uc_flags & RAWVER)
            {
               putchar ('V');
               i--;
            }
            else if (ucb.uc_flags & RAWCOMP)
            {
               putchar ('S');
               i--;
            }
            if (ucb.uc_flags & UDEBUG)
            {
               putchar ('D');
               i--;
            }
            if (ucb.uc_flags & ULISTEN)
            {
               putchar ('L');
               i--;
            }
            while (i-- > 0)
               putchar (' ');

            printf ("%04o ", ucb.uc_state);

            printf ("%15s ", inet_ntoa(ucb.uc_host.s_addr));

            printf ("%5u %5u %04X %2d %2d %2d %2d %2d\n",
                    (tcp ? tcb.t_fport : udp ? ucb.uc_udp.u_fport : 0),
                    (tcp ? tcb.t_lport : udp ? ucb.uc_udp.u_lport : 0),
                    (tcp ? ucb.uc_tcb : 0),
                    ucb.uc_ssize,
                    ucb.uc_rsize,
                    ucb.uc_snd, ucb.uc_rcv, (tcp ? tcb.t_state : 0));
         }

      }
      if (all) putchar ('\n');
   }

   if (rstats)
   {
      printf ("LOCAL NET       FOREIGN NET     GATEWAY         FLAGS\n");
      for (gp = gateway; gp < gatewayNGATE; gp++)
      {
         printf ("%15s ", inet_ntoa(nettoip(gp->g_lnet)));
         printf ("%15s ", inet_ntoa(nettoip(gp->g_fnet)));
         printf ("%15s ", inet_ntoa(gp->g_local.s_addr));
         printf ("%ld\n", gp->g_flags);
      }
      if (all) putchar ('\n');
   }

   if (istats)
   {
      printf ("NAME     FLAGS   MTU ADDR            OPKT OERR IPKT IERR\n");
               
      for (lp = netcb.n_ifcb_hd; lp != NULL; lp = lp->if_next)
      {
         memcpy (&lifcb, lp, sizeof (struct ifcb));
	 printf ("%8s ", lifcb.if_name);
	 i = 7;
         if (lifcb.if_avail)
         {
	    putchar ('U');
	    i--;
         }
         if (lifcb.if_error)
         {
	    putchar ('E');
	    i--;
         }
         if (lifcb.if_needinit)
         {
	    putchar ('I');
	    i--;
         }
         if (lifcb.if_active)
         {
	    putchar ('A');
	    i--;
         }
         if (lifcb.if_flush)
         {
	    putchar ('F');
	    i--;
         }
         if (lifcb.if_blocked)
         {
	    putchar ('B');
	    i--;
         }
         while (i-- > 0)
	    putchar (' ');
	 printf ("%4d ", lifcb.if_mtu);
	 printf ("%15s ", inet_ntoa(lifcb.if_addr.s_addr));
	 printf ("%4d ", lifcb.if_opkts);
	 printf ("%4d ", lifcb.if_oerrs);
	 printf ("%4d ", lifcb.if_ipkts);
	 printf ("%4d\n", lifcb.if_ierrs);
      }
      if (all) putchar ('\n');
   }

   if (mstats)
   {
      printf ("pages = %d, bufs = %d, free = %d, hiwat = %d, lowat = %d\n",
              netcb.n_pages,
              (netcb.n_pages << 3),
              netcb.n_bufs, netcb.n_hiwat, netcb.n_lowat);
   }

   if (pstats)
   {
      printf ("mem drops = %d\n", netstat.m_drops);
      printf (
      "ip badsums = %d, tcp badsums = %d, tcp rejects = %d, tcp unacked = %d\n",
          netstat.ip_badsum, netstat.t_badsum, netstat.t_badsegs,
          netstat.t_unack);
      printf (
      "icmp drops = %d, icmp badsums = %d, src quenches = %d, redirects = %d\n",
          netstat.ic_drops, netstat.ic_badsum, netstat.ic_quenches,
          netstat.ic_redirects);
      printf ("icmp echoes = %d, time exceeded = %d\n", netstat.ic_echoes,
              netstat.ic_timex);
   }
}
