#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include "inet.h"

#undef NULL

#include "net.h"
#include "con.h"
#include "udp.h"

static struct con con;

static char str[] = "Hello world!\n";
static char msg[512];
static int sd;

static void 
sigcatcher (sig)
{
   if (sd)
      netclose (sd);
}

main (argc, argv)
     char **argv;
{
   unsigned long ip_addr;
   char *bp;
   int lcnt, i;
   int len;
   int tcptest;
   int readreply;
   int port;

   ip_addr = 0;
   sd = 0;
   lcnt = 1;
   tcptest = 0;
   readreply = 0;
   len = 0;
   port = 4000;

   /*
    ** Scan off args 
    */

   for (i = 1; i < argc; i++)
   {
      bp = argv[i];

      if (*bp == '-')
      {
         for (bp++; *bp; bp++)
            switch (*bp)
            {
            case 'n':
               i++;
               lcnt = atoi (argv[i]);
               break;

            case 'r':
               readreply = 1;
               break;

            case 't':
               tcptest = 1;
               break;

            default:
	    USAGE:
               printf ("usage: nettest [-options] ip_addr [port]\n");
               printf (" options:\n");
               printf ("    -n number    - Number of messages to send\n");
               printf ("    -r           - Read reply\n");
               printf ("    -t           - Send over TCP\n");
               return (1);
            }
      }
      else if (len == 0)
      {
         ip_addr = inet_addr (argv[i]);
	 len++;
      }
      else if (len == 1)
      {
         port = atoi (argv[i]);
	 len++;
      }
      else 
      {
         goto USAGE;
      }
   }

   printf
      ("Start: ip_addr = %s, port = %d, reply = %d, tcp = %d, count = %d\n",
       inet_ntoa(ip_addr), port, readreply, tcptest, lcnt);

   if (ip_addr == 0)
      goto USAGE;

   memset (&con, 0, sizeof (struct con));
   signal (SIGINT, sigcatcher);

   if (!tcptest)
   {

      printf ("Open UDP socket\n");
      con.c_mode = CONUDP | CONACT | CONRAWCOMP;

      con.c_fcon.s_addr = ip_addr;
      con.c_lcon.s_addr = 0;

      con.c_fport = port;
      con.c_lport = port;
      con.c_timeo = 60;

      sd = netopen (&con);
      if (errno)
         goto BADNESS;
      printf ("socket# %d\n", sd);

      for (i = 0; i < lcnt; i++)
      {
         printf ("Write to socket\n");
         if (netwrite (sd, str, strlen (str)) < 0)
            goto BADNESS;

         if (readreply)
         {
            printf ("Read from socket\n");
            if ((len = netread (sd, msg, sizeof (msg))) > 0) {
               printf ("Received %d bytes\n", len);
               printf ("Message: %s\n", msg + sizeof (struct udp));
            }
            else {
               printf ("NO REPLY\n");
               if (errno)
                  goto BADNESS;
	    }
         }
      }
   }

   else
   {
      printf ("Open TCP client\n");

      con.c_mode = CONTCP | CONACT;
      con.c_fcon.s_addr = ip_addr;
      con.c_lcon.s_addr = 0;
      con.c_fport = port;
      con.c_lport = 5000;
      con.c_timeo = 60;

      if ((sd = netopen (&con)) < 0)
      {
         printf ("client open failed: error = %d\n", errno);
         exit (1);
      }
      printf ("client socket# %d\n", sd);

      for (i = 0; i < lcnt; i++)
      {
         printf ("Write to client socket\n");
         if (netwrite (sd, str, strlen (str)) < 0)
	    goto BADNESS;

         if (readreply)
         {
            printf ("Read from server socket\n");
            if ((len = netread (sd, msg, sizeof (msg))) > 0) {
               printf ("Received %d bytes\n", len);
               printf ("Message: %s\n", msg);
            }
            else {
               printf ("NO REPLY\n");
               if (errno)
                  goto BADNESS;
	    }
         }
      }
   }
   netclose (sd);
   return 0;

 BADNESS:
   printf ("ERROR = %d\n", errno);
   netclose (sd);
   return 1;
}
