
#include "param.h"

#ifdef TI_CTX

/* Initial state of kernel and user mode page maps. For the kernel
 * map only pages 12, 13 and 14 change. Page 0 of the user map is
 * always 0x80 (physical page 0, write protected).
 */
char sysmap[16] = {
   0x00, 0x08, 0x01, 0x09, 0x02, 0x0a, 0x03, 0x0b,
   0x04, 0x0c, 0x05, 0x0d, 0x06, 0x0e, 0x07, 0x0f
};
char usrmap[16] = {
   0x00, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
   0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x0f
};

#else
unsigned int sysmap[6] = {
   0x2000,	/* L1 57K - kernel */
   0x0000,	/* B1 */
   0x1000,	/* L2 4K - _u/swapper */ 
   0x0000,	/* B2  changes with slot */
   0x0000,	/* L3 4K  - WS + TILINE */
   0x0000	/* B3 */
};
unsigned int usrmap[6] = {
   0xF001,	/* L1 4K - _u */
   0x0000,	/* B1 */
   0x1000,	/* L2 57K - user code */ 
   0x0000,	/* B2  changes with slot */
   0x0802,	/* L3 4K  - WS - NO TILINE */
   0x0000	/* B3 */
};

#endif

/*
 * CORE
 */

#define FULL	1
#define FREE	0

char coremap[CMAPSIZ];	/* space for core allocation */
extern int maxslot;

/* Find a free slot in physical memory and return its slot
 * number. Return zero if no free slot could be found.
 */
int
memget()
{
	register int i;
	
	for(i=1; i<maxslot; i++) {
		if(coremap[i]==FREE) break;
	}
	if(i<maxslot) {
		coremap[i] = FULL;
		return i;
	}
	return 0;
}

/* Return memory slot to the free pool */
void
memfree(slot)
{
	coremap[slot] = FREE;
}

/*
 * SWAP
 */

char	swapmap[3*SMAPSIZ];	/* space for swap allocation */

#define PROCSIZ		120	/* 120 blocks = 60KB */
extern daddr_t swplo;		/* start of swap space */

/* Find a free slot in swap space and return the disk address
 * of its first block. Return zero if no free slot could be
 * found. If 'zombie' is non zero space for single block is
 * returned, else space for 60KB is returned.
 */
daddr_t
swpget(zombie)
{
	register int i;

	if(zombie) {
		for(i=SMAPSIZ; i<3*SMAPSIZ; i++) {
			if(swapmap[i]==FREE) break;
		}
		if(i<3*SMAPSIZ) {
			swapmap[i] = FULL;
			return swplo + (SMAPSIZ*PROCSIZ) + i;
		}
		return 0;		
	}
	else {
		for(i=0; i<SMAPSIZ; i++) {
			if(swapmap[i]==FREE) break;
		}
		if(i<SMAPSIZ) {
			swapmap[i] = FULL;
			return swplo + (i*PROCSIZ);
		}
		return 0;
	}
}

/* Return swap slot to the free pool */
void
swpfree(slot)
{
	register daddr_t i;
	
	i = slot - swplo;
	if (i<SMAPSIZ*PROCSIZ )
		i /= PROCSIZ;
	else
		i -= SMAPSIZ*PROCSIZ;
	swapmap[i] = FREE;
}

/* Return the physical page number for address 'a' in the
 * virtual memory space of the process in slot 's'. For the benefit
 * of swapping, virtual page zero maps to the process' u page.
 */
unsigned int
physpage(a, s)
	char *a;
{
	register int i;

	i = (unsigned)a>>12;
	return (i | s<<4);
}
